import numpy as np
import matplotlib.pyplot as plt
from cmath import exp
from math import sin,  cos

def evolve(l, j, hx, v, dt, n):
    """ Evolve the state v in time by n * dt """
    dim = 1 << l
    def tstep_diag(x, dt):
        # Apply diagonal part
        for s in range(dim):
            jtotal = 0.
            for r in range(l-1):
                jtotal += -j[r] if ((s >> r)^(s >> (r+1)))&1 else j[r]
            x[s] = exp(-1.j*jtotal*dt)*x[s]
        return x
    def tstep_transv(x, dt):
        # Apply transverse part succesively for each site
        y =  np.zeros_like(x)
        for r in range(l):
            cos_v = cos(dt*hx[r])
            sin_v = sin(dt*hx[r])
            # diagonal
            for s in range(dim):
                y[s] = cos_v * x[s]
            # off-diagonal
            for s in range(dim):
                y[s^(1<<r)] += 1.j*sin_v * x[s]
            # swap vectors
            tmp = x
            x   = y
            y   = tmp
        return x
    v = tstep_diag(v, dt/2.)
    for i in range(int(n-1)):
        v = tstep_transv(v, dt)
        v = tstep_diag(v, dt)
    v = tstep_transv(v, dt)
    v = tstep_diag(v, dt/2.)
    return v
    
def magnetization(v, l):
    """ Measure the magnetization per site for the state v """
    dim = 1 << l
    m = np.zeros((l))
    for r in range(l):
        for s in range(dim):
            m[r] += abs(v[s])**2 * 2. * ( bool(s&(1 << r)) -0.5)
    return m


l   =  9
dim =  1 << l
j   =  1.00*np.ones((l-1)) # Ising coupling
hx  =  0.40*np.ones((l)) # transverse field strength Gamma

# Starting configuration
ind = int('0000001000000',base=2) # specify spin configuration: 0 = down, 1 = up
v   = np.zeros((dim),dtype=complex)
v[ind] = 1.

# Parameters for time evolution
tend   = 20. # duration of the time evolution
tmeas  = 0.5 # time after which a measurment is performed
dt     = 0.1 # time step used for the evolution
nmeas  = int((tend)/tmeas) # number of measurements
nstep  = int((tmeas)/dt) # number of steps between measurments

# measure magnetization
m=np.zeros((nmeas,l));
m[0,:] = magnetization(v, l)

# Do time evolution
for t in range(1,nmeas):
    print(t*tmeas)
    v=evolve(l,j,hx,v,dt,nstep)
    m[t,:] = magnetization(v, l)

# plot magnetization
plt.pcolor(np.array(range(l+1)),np.linspace(0.,tend,nmeas),m,
           vmin=-1.0, vmax=1.0)
plt.xlabel('site')
plt.ylabel('time')
plt.axis([0,l,0,tend])
plt.title("Magnetisation, J = "+str(j[0])+" , hx = "+str(hx[0]))
plt.colorbar()
plt.show()
